//
//  Intent+Link.swift
//  StripePaymentSheet
//
//  Created by Ramon Torres on 11/3/21.
//  Copyright © 2021 Stripe, Inc. All rights reserved.
//

@_spi(STP) import StripePayments

extension Intent {
    var supportsLink: Bool {
        return recommendedPaymentMethodTypes.contains(.link) || linkPassthroughModeEnabled
    }

    var supportsLinkCard: Bool {
        return supportsLink && (linkFundingSources?.contains(.card) ?? false) || linkPassthroughModeEnabled
    }

    var onlySupportsLinkBank: Bool {
        return supportsLink && (linkFundingSources == [.bankAccount])
    }

    var callToAction: ConfirmButton.CallToActionType {
        switch self {
        case .paymentIntent(let paymentIntent):
            return .pay(amount: paymentIntent.amount, currency: paymentIntent.currency)
        case .setupIntent:
            return .setup
        case .deferredIntent(_, let intentConfig):
            switch intentConfig.mode {
            case .payment(let amount, let currency, _, _):
                return .pay(amount: amount, currency: currency)
            case .setup:
                return .setup
            }
        }
    }

    var linkPassthroughModeEnabled: Bool {
        switch self {
        case .paymentIntent(let paymentIntent):
            return paymentIntent.linkSettings?.passthroughModeEnabled ?? false
        case .setupIntent(let setupIntent):
            return setupIntent.linkSettings?.passthroughModeEnabled ?? false
        case .deferredIntent(let elementsSession, _):
            return elementsSession.linkSettings?.passthroughModeEnabled ?? false
        }
    }

    var linkFundingSources: Set<LinkSettings.FundingSource>? {
        switch self {
        case .paymentIntent(let paymentIntent):
            return paymentIntent.linkSettings?.fundingSources
        case .setupIntent(let setupIntent):
            return setupIntent.linkSettings?.fundingSources
        case .deferredIntent(let elementsSession, _):
            return elementsSession.linkSettings?.fundingSources
        }
    }

    var linkPopupWebviewOption: LinkSettings.PopupWebviewOption {
        return {
            switch self {
            case .paymentIntent(let paymentIntent):
                return paymentIntent.linkSettings?.popupWebviewOption
            case .setupIntent(let setupIntent):
                return setupIntent.linkSettings?.popupWebviewOption
            case .deferredIntent(let elementsSession, _):
                return elementsSession.linkSettings?.popupWebviewOption
            }
        }() ?? .shared
    }

    var countryCode: String? {
        switch self {
        case .paymentIntent(let paymentIntent):
            return paymentIntent.countryCode
        case .setupIntent(let setupIntent):
            return setupIntent.countryCode
        case .deferredIntent(let elementsSession, _):
            return elementsSession.countryCode
        }
    }

    var merchantCountryCode: String? {
        switch self {
        case .paymentIntent(let paymentIntent):
            return paymentIntent.merchantCountryCode
        case .setupIntent(let setupIntent):
            return setupIntent.merchantCountryCode
        case .deferredIntent(let elementsSession, _):
            return elementsSession.merchantCountryCode
        }
    }
}
